/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/* smb-method.h - VFS modules for SMB
 *
 *  Copyright (C) 2001 Red Hat Inc,
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Author: Alexander Larsson <alexl@redhat.com>
 *         Tambet Ingo <tambet@ximian.com>
 *
 * Some inspiration comes from the smbfs implementation in midnight commander
 * by Wayne Roberts <wroberts1@home.com>. 
 */


#include "config.h"
#include "gnome-vfs-extras-i18n.h"

#include <errno.h>
#include <string.h>
#include <stdio.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime.h>
#include <libgnomevfs/gnome-vfs-mime-utils.h>
#include <glib/gutils.h>

#include <libgnomevfs/gnome-vfs-method.h>
#include <libgnomevfs/gnome-vfs-module.h>
#include <libgnomevfs/gnome-vfs-module-shared.h>
#include <libgnomevfs/gnome-vfs-module-callback-module-api.h>
#include <libgnomevfs/gnome-vfs-standard-callbacks.h>

#include "smb-method.h"

#include <libsmbclient.h>

static GMutex *samba_lock = NULL;


#undef DEBUG_SMB_ENABLE
#undef DEBUG_SMB_LOCKS
#undef DEBUG_SMB_2

#ifdef DEBUG_SMB_ENABLE
#define DEBUG_SMB(x) g_warning x
#else
#define DEBUG_SMB(x)
#endif

#ifdef DEBUG_SMB_LOCKS
#define LOCK_SAMBA() 	{g_mutex_lock (samba_lock); g_print ("LOCK %s\n", G_GNUC_PRETTY_FUNCTION);}
#define UNLOCK_SAMBA() 	{g_print ("UNLOCK %s\n", G_GNUC_PRETTY_FUNCTION); g_mutex_unlock (samba_lock);}
#else
#define LOCK_SAMBA() 	g_mutex_lock (samba_lock)
#define UNLOCK_SAMBA() 	g_mutex_unlock (samba_lock)
#endif

#ifdef DEBUG_SMB_2
#define DEBUG_SMB2(x) g_print x
#else
#define DEBUG_SMB2(x)
#endif

typedef struct
{
  GnomeVFSURI *uri;
  gint fd;
} SmbFileHandle;


static GnomeVFSResult
do_open (GnomeVFSMethod * method,
	 GnomeVFSMethodHandle ** method_handle,
	 GnomeVFSURI * uri, GnomeVFSOpenMode mode, GnomeVFSContext * context)
{
  gint flags;
  gint fd;
  SmbFileHandle *handle;
  gchar *full_name;

  if ((mode & GNOME_VFS_OPEN_READ) && (mode & GNOME_VFS_OPEN_WRITE))
    {
      flags = O_RDWR;
    }
  else if (mode & GNOME_VFS_OPEN_READ)
    {
      flags = O_RDONLY;
    }
  else if (mode & GNOME_VFS_OPEN_WRITE)
    {
      flags = O_WRONLY;
    }
  else
    {
      return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
    }

  full_name = gnome_vfs_uri_to_string (uri, 0);

  LOCK_SAMBA ();
  DEBUG_SMB (("do_open() %s mode %d\n", full_name, mode));

  fd = smbc_open (full_name, flags, 0666);

  if (fd < 0)
    {
      UNLOCK_SAMBA ();
      DEBUG_SMB (("Samba has returned an error %s", strerror (errno)));
      g_free (full_name);
      return gnome_vfs_result_from_errno ();
    }
  UNLOCK_SAMBA ();

  handle = g_new (SmbFileHandle, 1);
  handle->fd = fd;
  handle->uri = gnome_vfs_uri_ref (uri);
  *method_handle = (GnomeVFSMethodHandle *) handle;
  g_free (full_name);

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_close (GnomeVFSMethod * method,
	  GnomeVFSMethodHandle * method_handle, GnomeVFSContext * context)
{

  LOCK_SAMBA ();
  DEBUG_SMB (("do_close() %p\n", method_handle));

  if (smbc_close (((SmbFileHandle *) method_handle)->fd) < 0)
    {
      UNLOCK_SAMBA ();
      DEBUG_SMB (("Samba has returned an error %s", strerror (errno)));
      return gnome_vfs_result_from_errno ();
    }


  g_free (method_handle);
  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_read (GnomeVFSMethod * method,
	 GnomeVFSMethodHandle * method_handle,
	 gpointer buffer,
	 GnomeVFSFileSize num_bytes,
	 GnomeVFSFileSize * bytes_read, GnomeVFSContext * context)
{
  SmbFileHandle *file_handle;
  gint read_val;


  g_return_val_if_fail (method_handle != NULL, GNOME_VFS_ERROR_INTERNAL);
  LOCK_SAMBA ();
  DEBUG_SMB (("do_read() %p\n", method_handle));

  file_handle = (SmbFileHandle *) method_handle;

  do
    {
      read_val = smbc_read (file_handle->fd, buffer, num_bytes);
    }
  while (read_val == -1
	 && errno == EINTR
	 && !gnome_vfs_context_check_cancellation (context));

  if (read_val == -1)
    {
      *bytes_read = 0;
      UNLOCK_SAMBA ();
      return gnome_vfs_result_from_errno ();
    }
  else
    {
      *bytes_read = read_val;

      /* Getting 0 from read() means EOF! */
      if (read_val == 0)
	{
	  UNLOCK_SAMBA ();
	  return GNOME_VFS_ERROR_EOF;
	}
    }
  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_write (GnomeVFSMethod * method,
	  GnomeVFSMethodHandle * method_handle,
	  gconstpointer buffer,
	  GnomeVFSFileSize num_bytes,
	  GnomeVFSFileSize * bytes_written, GnomeVFSContext * context)
{
  SmbFileHandle *file_handle;
  gint write_val;

  g_return_val_if_fail (method_handle != NULL, GNOME_VFS_ERROR_INTERNAL);

  LOCK_SAMBA ();
  DEBUG_SMB (("do_write() %p\n", method_handle));
  file_handle = (SmbFileHandle *) method_handle;

  do
    write_val = smbc_write (file_handle->fd, buffer, num_bytes);
  while (write_val == -1
	 && errno == EINTR
	 && !gnome_vfs_context_check_cancellation (context));

  if (write_val == -1)
    {
      *bytes_written = 0;
      UNLOCK_SAMBA ();
      return gnome_vfs_result_from_errno ();
    }
  else
    {
      *bytes_written = write_val;
      UNLOCK_SAMBA ();
      return GNOME_VFS_OK;
    }

  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_create (GnomeVFSMethod * method,
	   GnomeVFSMethodHandle ** method_handle,
	   GnomeVFSURI * uri,
	   GnomeVFSOpenMode mode,
	   gboolean exclusive, guint perm, GnomeVFSContext * context)
{
  gint flags;
  gint fd;
  SmbFileHandle *handle;
  gchar *full_name;

  full_name = gnome_vfs_uri_to_string (uri, 0);

  if ((mode & GNOME_VFS_OPEN_READ) && (mode & GNOME_VFS_OPEN_WRITE))
    {
      flags = O_RDWR;
    }
  else if (mode & GNOME_VFS_OPEN_READ)
    {
      flags = O_RDONLY;
    }
  else if (mode & GNOME_VFS_OPEN_WRITE)
    {
      flags = O_WRONLY;
    }
  else
    {
      return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
    }

  flags |= O_CREAT;
  if (exclusive)
    {
      flags |= O_EXCL;
    }

  LOCK_SAMBA ();

  DEBUG_SMB (("do_create() %s mode %d\n", full_name, mode));
  fd = smbc_creat (full_name, flags);
  if (fd < 0)
    {
      UNLOCK_SAMBA ();
      g_free (full_name);
      return gnome_vfs_result_from_errno ();
    }


  UNLOCK_SAMBA ();

  handle = g_new (SmbFileHandle, 1);
  handle->fd = fd;
  handle->uri = gnome_vfs_uri_ref (uri);
  *method_handle = (GnomeVFSMethodHandle *) handle;

  g_free (full_name);

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_get_file_info (GnomeVFSMethod * method,
		  GnomeVFSURI * uri,
		  GnomeVFSFileInfo * file_info,
		  GnomeVFSFileInfoOptions options, GnomeVFSContext * context)
{
  struct stat stat_info;
  gchar *escaped_base;
  gchar *full_name;

  full_name = gnome_vfs_uri_to_string (uri, 0);

  file_info->flags = GNOME_VFS_FILE_FLAGS_NONE;

  escaped_base = gnome_vfs_uri_extract_short_path_name (uri);
  file_info->name =
    gnome_vfs_unescape_string (escaped_base, G_DIR_SEPARATOR_S);
  g_free (escaped_base);


  LOCK_SAMBA ();
  DEBUG_SMB2 (("do_get_file_info() %s\n", gnome_vfs_uri_to_string (uri, 0)));

  if (smbc_stat (full_name, &stat_info) < 0)
    {
      g_free (full_name);
      UNLOCK_SAMBA ();
      return gnome_vfs_result_from_errno ();
    }

  gnome_vfs_stat_to_file_info (file_info, &stat_info);

  if (S_ISDIR (stat_info.st_mode))
    file_info->mime_type = g_strdup ("x-directory/normal");
  else
    file_info->mime_type =
      g_strdup (gnome_vfs_mime_type_from_name_or_default
		(file_info->name, GNOME_VFS_MIME_TYPE_UNKNOWN));
  file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;

  UNLOCK_SAMBA ();
  g_free (full_name);

  return GNOME_VFS_OK;
}

static gboolean
do_is_local (GnomeVFSMethod * method, const GnomeVFSURI * uri)
{
  return FALSE;
}



static GnomeVFSResult
do_open_directory (GnomeVFSMethod * method,
		   GnomeVFSMethodHandle ** method_handle,
		   GnomeVFSURI * uri,
		   GnomeVFSFileInfoOptions options, GnomeVFSContext * context)
{
  gint fd;
  SmbFileHandle *handle;
  gchar *full_name;

  LOCK_SAMBA ();
  DEBUG_SMB (("do_opendir() %s\n", gnome_vfs_uri_to_string (uri, 0)));

  full_name = gnome_vfs_uri_to_string (uri, 0);
  fd = smbc_opendir (full_name);
  if (fd < 0)
    {
      UNLOCK_SAMBA ();
      g_free (full_name);
      return gnome_vfs_result_from_errno ();
    }

  UNLOCK_SAMBA ();

  handle = g_new (SmbFileHandle, 1);
  handle->fd = fd;
  handle->uri = gnome_vfs_uri_ref (uri);
  *method_handle = (GnomeVFSMethodHandle *) handle;

  g_free (full_name);

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_close_directory (GnomeVFSMethod * method,
		    GnomeVFSMethodHandle * method_handle,
		    GnomeVFSContext * context)
{
  LOCK_SAMBA ();
  DEBUG_SMB2 (("do_close_directory() %p\n", handle));


  if (smbc_closedir (((SmbFileHandle *) method_handle)->fd) < 0)
    {
      UNLOCK_SAMBA ();
      return gnome_vfs_result_from_errno ();
    }


  g_free (method_handle);
  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_read_directory (GnomeVFSMethod * method,
		   GnomeVFSMethodHandle * method_handle,
		   GnomeVFSFileInfo * file_info, GnomeVFSContext * context)
{
  struct smbc_dirent *dirent;
  /*
     struct stat stat_info;
   */
  SmbFileHandle *handle;

  handle = (SmbFileHandle *) method_handle;

  DEBUG_SMB2 (("do_read_directory()\n"));


  dirent = smbc_readdir (handle->fd);

  if (dirent == NULL && errno != 0)
    {
      return gnome_vfs_result_from_errno ();
    }
  if (dirent == NULL)
    {
      return GNOME_VFS_ERROR_EOF;
    }

  file_info->name = g_strdup (dirent->name);
  switch (dirent->smbc_type)
    {
    case SMBC_WORKGROUP:
    case SMBC_SERVER:
      file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
      file_info->mime_type = g_strdup ("application/x-gnome-app-info");

      file_info->valid_fields |=
	GNOME_VFS_FILE_INFO_FIELDS_TYPE |
	GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;

      break;
    case SMBC_FILE_SHARE:
    case SMBC_DIR:
      file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
      file_info->mime_type = g_strdup ("x-directory/smb-share");

      file_info->valid_fields |=
	GNOME_VFS_FILE_INFO_FIELDS_TYPE |
	GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;

      break;
    default:
      file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
      file_info->mime_type =
	g_strdup (gnome_vfs_mime_type_from_name_or_default (file_info->name,
							    GNOME_VFS_MIME_TYPE_UNKNOWN));

      file_info->valid_fields |=
	GNOME_VFS_FILE_INFO_FIELDS_TYPE |
	GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
    }


/*        if (smbc_stat(file_info->name, &stat_info)<0)
	  {
      	    gnome_vfs_stat_to_file_info (file_info, &stat_info);
	    return gnome_vfs_result_from_errno();
	  }
*/
  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_seek (GnomeVFSMethod * method,
	 GnomeVFSMethodHandle * method_handle,
	 GnomeVFSSeekPosition whence,
	 GnomeVFSFileOffset offset, GnomeVFSContext * context)
{
  SmbFileHandle *file_handle;
  gint lseek_whence;

  file_handle = (SmbFileHandle *) method_handle;

  DEBUG_SMB (("do_seek() %d from %d\n", (guint) offset, whence));

  switch (whence)
    {
    case GNOME_VFS_SEEK_START:
      lseek_whence = SEEK_SET;
    case GNOME_VFS_SEEK_CURRENT:
      lseek_whence = SEEK_CUR;
    case GNOME_VFS_SEEK_END:
      lseek_whence = SEEK_END;
    default:
      g_warning (_("Unknown GnomeVFSSeekPosition %d"), whence);
      lseek_whence = SEEK_SET;	/* bogus */
    }

  LOCK_SAMBA ();

  if (smbc_lseek (file_handle->fd, offset, lseek_whence) == -1)
    {

      UNLOCK_SAMBA ();

      if (errno == ESPIPE)
	return GNOME_VFS_ERROR_NOT_SUPPORTED;
      else
	return gnome_vfs_result_from_errno ();
    }

  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_tell (GnomeVFSMethod * method,
	 GnomeVFSMethodHandle * method_handle,
	 GnomeVFSFileOffset * offset_return)
{
  SmbFileHandle *file_handle;
  glong offset;

  file_handle = (SmbFileHandle *) method_handle;

  LOCK_SAMBA ();
  offset = smbc_lseek (file_handle->fd, 0, SEEK_CUR);
  UNLOCK_SAMBA ();

  if (offset == -1)
    {
      if (errno == ESPIPE)
	return GNOME_VFS_ERROR_NOT_SUPPORTED;
      else
	return gnome_vfs_result_from_errno ();
    }

  *offset_return = offset;
  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_unlink (GnomeVFSMethod * method,
	   GnomeVFSURI * uri, GnomeVFSContext * context)
{
  gchar *full_name;
  gint retval;

  full_name = gnome_vfs_uri_to_string (uri, 0);
  LOCK_SAMBA ();
  DEBUG_SMB (("do_unlink() %s\n", full_name));


  retval = smbc_unlink (full_name);
  UNLOCK_SAMBA ();
  g_free (full_name);

  if (retval != 0)
    {
      return gnome_vfs_result_from_errno ();
    }

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_check_same_fs (GnomeVFSMethod * method,
		  GnomeVFSURI * a,
		  GnomeVFSURI * b,
		  gboolean * same_fs_return, GnomeVFSContext * context)
{
  char *server1;
  char *server2;
  char *path1;
  char *path2;
  char *p1, *p2;

  server1 = gnome_vfs_unescape_string (gnome_vfs_uri_get_host_name (a), NULL);
  server2 = gnome_vfs_unescape_string (gnome_vfs_uri_get_host_name (b), NULL);
  path1 = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (a), NULL);
  path2 = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (b), NULL);

  DEBUG_SMB (("do_check_same_fs( %s, %s, %s, %s)\n", server1, server2, path1, path2));

  if (!server1 || !server2 || !path1 || !path2 ||
      (strcmp (server1, server2) != 0))
    {
      g_free (server1);
      g_free (server2);
      g_free (path1);
      g_free (path2);
      *same_fs_return = FALSE;
      return GNOME_VFS_OK;
    }

  p1 = path1;
  p2 = path2;
  if (*p1 == '/')
    {
      p1++;
    }
  if (*p2 == '/')
    {
      p2++;
    }

  /* Make sure both URIs are on the same share: */
  while (*p1 && *p2 && *p1 == *p2 && *p1 != '/')
    {
      p1++;
      p2++;
    }
  if (*p1 == 0 || *p2 == 0 || *p1 != *p2)
    {
      *same_fs_return = FALSE;
    }
  else
    {
      *same_fs_return = TRUE;
    }

  g_free (server1);
  g_free (server2);
  g_free (path1);
  g_free (path2);

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_move (GnomeVFSMethod * method,
	 GnomeVFSURI * old_uri,
	 GnomeVFSURI * new_uri,
	 gboolean force_replace, GnomeVFSContext * context)
{

  LOCK_SAMBA ();
  DEBUG_SMB (("do_move() %s %s\n",
	      gnome_vfs_uri_to_string (old_uri, 0),
	      gnome_vfs_uri_to_string (new_uri, 0)));
  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}


static GnomeVFSResult
do_make_directory (GnomeVFSMethod * method,
		   GnomeVFSURI * uri, guint perm, GnomeVFSContext * context)
{
  gchar *full_name;
  gint retval;

  full_name = gnome_vfs_uri_to_string (uri, 0);
  LOCK_SAMBA ();

  DEBUG_SMB (("do_make_dir() %s\n", full_name));

  retval = smbc_mkdir (full_name, perm);

  g_free (full_name);
  UNLOCK_SAMBA ();

  if (retval != 0)
    {
      return gnome_vfs_result_from_errno ();
    }

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_remove_directory (GnomeVFSMethod * method,
		     GnomeVFSURI * uri, GnomeVFSContext * context)
{
  gchar *full_name;
  gint retval;

  full_name = gnome_vfs_uri_to_string (uri, 0);
  LOCK_SAMBA ();

  DEBUG_SMB (("do_remove_dir() %s\n", full_name));

  retval = smbc_rmdir (full_name);

  g_free (full_name);
  UNLOCK_SAMBA ();

  if (retval != 0)
    {
      return gnome_vfs_result_from_errno ();
    }

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_set_file_info (GnomeVFSMethod * method,
		  GnomeVFSURI * uri,
		  const GnomeVFSFileInfo * info,
		  GnomeVFSSetFileInfoMask mask, GnomeVFSContext * context)
{

  DEBUG_SMB (("do_set_file_info mask: %d\n", mask));
  LOCK_SAMBA ();
  UNLOCK_SAMBA ();

  return GNOME_VFS_OK;
}

static GnomeVFSMethod method = {
  sizeof (GnomeVFSMethod),
  do_open,
  do_create,
  do_close,
  do_read,
  do_write,
  do_seek,
  do_tell,
  NULL,				/* do_truncate_handle, */
  do_open_directory,
  do_close_directory,
  do_read_directory,
  do_get_file_info,
  NULL,				/* do_get_file_info_from_handle, */
  do_is_local,
  do_make_directory,
  do_remove_directory,
  do_move,
  do_unlink,
  do_check_same_fs,
  do_set_file_info,
  NULL,				/* do_truncate, */
  NULL,				/* do_find_directory, */
  NULL				/* do_create_symbolic_link */
};

static void
vfs_module_auth_fn (const char *server, const char *share,
		    char *workgroup, int wgmaxlen, char *username,
		    int unmaxlen, char *password, int pwmaxlen)
{
/*  GnomeVFSModuleCallbackAuthenticationIn in_args;
  GnomeVFSModuleCallbackAuthenticationOut out_args;
	
  memset (&in_args, 0, sizeof (in_args));			
  memset (&out_args, 0, sizeof (out_args));
	
  in_args.uri = g_strconcat ("Connecting to ", server, "/", share, NULL);			

  gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_AUTHENTICATION,
 				    &in_args, sizeof (in_args), 
				    &out_args, sizeof (out_args));	
  g_free(in_args.uri);
						
  strncpy (username, out_args.username, unmaxlen);
  strncpy (password, out_args.password, pwmaxlen);*/
}


GnomeVFSMethod *
vfs_module_init (const char *method_name, const char *args)
{

#ifdef ENABLE_NLS
  bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
  samba_lock = g_mutex_new ();

  smbc_init (vfs_module_auth_fn, 0);

  return &method;
}

void
vfs_module_shutdown (GnomeVFSMethod * method)
{
  DEBUG_SMB (("<-- smb module shutdown called -->\n"));
}
